import 'dart:convert';
import 'dart:typed_data';

import '../utils/monotonic_id.dart';

import '../logger/console_logger.dart';
import '../logger/logger.dart';
import '../migratorydata_client.dart';
import '../migratorydata_listener.dart';
import '../migratorydata_log_listener.dart';
import '../migratorydata_message.dart';
import '../protocol/push_protocol.dart';
import '../protocol/push_protocol_transport_codec.dart';
import 'cluster_info.dart';
import 'configuration.dart';
import 'connection.dart';
import 'listener_notifier.dart';
import 'scheduler.dart';
import 'util.dart';
import 'version.dart';

class ClientImpl {
  Logger logger = ConsoleLogger.DEFAULT;

  bool running = false;

  Uint8List SESSION_TYPE = PushProtocolTransportCodec.instance
      .integerToBytes(SessionType.MOBILE.code);
  Uint8List USER_AGENT = PushProtocolTransportCodec.instance.escape(
      Uint8List.fromList(utf8.encode(
          "MigratoryDataClient/v6.0 DART, version: ${Version.VERSION}")));

  late ListenerNotifier listenerNotifier;

  late List<String> servers;
  late Configuration configuration;
  late Connection connection;
  late MigratoryDataListener migratoryDataListener;

  ClientState clientState = ClientState();

  ClientImpl() {
    configuration = Configuration(SESSION_TYPE, USER_AGENT, Version.VERSION);
  }

  void connect() {
    if (running) {
      throw Exception("Error: connect() - already connected");
    }
    if (servers.isEmpty) {
      throw Exception(
          "Error: connect() - no server to connect to; use setServers() to specify one or more servers");
    }
    if (migratoryDataListener == null) {
      throw Exception(
          "Error: connect() - no listener set; use setListener() to specify a listener handler");
    }
    running = true;

    ClusterInfo cluster = ClusterInfo(servers, configuration.isEncryption());

    listenerNotifier = ListenerNotifier(migratoryDataListener);
    Scheduler scheduler = Scheduler(configuration, clientState);
    connection = Connection(listenerNotifier, scheduler, cluster, configuration,
        clientState, logger);

    Map<String, SubjectContext> subjects = configuration.getSubjects();
    for (var s in subjects.entries) {
      connection.subscribe([s.key], s.value.history);
    }

    connection.connect();
  }

  void setServers(List<String> servers) {
    if (running) {
      throw Exception(
          "Error: setServers() - already connected; use this method before connect()");
    }
    if (servers.isEmpty) {
      throw Exception(
          "Error: setServers() - empty list argument; expected a list of addresses");
    }
    logger.info(
        "Setting servers to connect to: $servers, using API version: ${configuration.getVersion()}");
    this.servers = servers;
  }

  MigratoryDataListener getListener() {
    return migratoryDataListener;
  }

  setListener(MigratoryDataListener listener) {
    if (running) {
      throw Exception(
          "Error: setListener() - already connected; use this method before connect()");
    }
    migratoryDataListener = listener;
  }

  setLogListener(
      MigratoryDataLogListener logListener, MigratoryDataLogLevel logLevel) {
    if (running) {
      throw Exception(
          "Error: setLogListener() - already connected; use this method before connect()");
    }
    logger.setMigratoryDataLogListener(logListener, logLevel);
  }

  // operations

  subscribe(List<String> subjects, int history) {
    if (subjects.isEmpty || !Util.areSubjectsValid(subjects)) {
      throw Exception(
          "Error: subscribe() - invalid argument; expected a list of valid topics");
    }
    if (history < 0) {
      throw Exception(
          "Error: subscribeWithHistory() - invalid second argument; expected an integer higher or equal to 0");
    }
    logger.info("Subscribing to: $subjects");
    configuration.addSubject(subjects, history);
    if (running) {
      connection.subscribe(subjects, history);
    }
  }

  unsubscribe(List<String> subjects) {
    if (subjects.isEmpty || !Util.areSubjectsValid(subjects)) {
      throw Exception(
          "Error: unsubscribe() - invalid argument; expected a list of valid topics");
    }
    configuration.removeSubjects(subjects);
    logger.info("Unsubscribing from: $subjects");
    if (running) {
      connection.unsubscribe(subjects);
    }
  }

  pause() {
    if (!running) {
      throw Exception(
          "Error: pause() - not connected; use this method after connect()");
    }
    logger.info("migratorydata client calls pause");
    connection.pause();
  }

  resume() {
    if (!running) {
      throw Exception(
          "Error: resume() - not connected; use this method after connect()");
    }
    logger.info("migratorydata client calls resume");
    connection.resume();
  }

  publish(MigratoryDataMessage message) {
    if (!running) {
      throw Exception(
          "Error: publish() - not connected; use this method after connect()");
    }
    String subject = message.getSubject();
    Uint8List content = message.getContent();
    if (subject.isEmpty || !Util.isPushSubjectValid(subject)) {
      throw Exception(
          "Error: publish() - invalid argument; expected a valid message with a non-empty topic: $subject");
    }
    if (content.isEmpty) {
      throw Exception(
          "Error: publish() - invalid argument; expected a valid message with a non-empty content");
    }
    connection.publish(message);
  }

  disconnect() {
    logger.info("Disconnect from push server and release all resources.");
    if (!running) {
      return;
    }
    running = false;
    connection.dispose();
  }

  List<String> getServers() {
    return servers;
  }

  List<String> getSubjects() {
    return configuration.getSubjects().keys.toList();
  }

  // configuration

  notifyAfterFailedConnectionAttempts(int n) {
    if (running) {
      throw Exception(
          "Error: notifyAfterFailedConnectionAttempts() - already connected; use this method before connect()");
    }
    if (n < 1) {
      throw Exception(
          "Error: notifyAfterFailedConnectionAttempts() - invalid argument; expected a positive integer");
    }
    configuration.setServersDownBeforeNotify(n);
    logger.info(
        "Configuring the number of failed connection attempts before sending a notification: $n");
  }

  setEncryption(bool b) {
    if (running) {
      throw Exception(
          "Error: setEncryption() - already connected; use this method before connect()");
    }
    configuration.setEncryption(b);
    logger.info("Configuring encryption to: $b");
  }

  setEntitlementToken(String token) {
    if (token.isEmpty) {
      throw Exception(
          "Error: setEntitlementToken() - invalid token, null or empty.");
    }
    configuration.setEntitlementToken(token);
    logger.info("Configuring entitlement token: $token");
    if (running) {
      connection.sendAuthorizationToken(token);
    }
  }

  setQuickReconnectMaxRetries(int number) {
    if (running) {
      throw Exception(
          "Error: setQuickReconnectMaxRetries() - already connected; use this method before connect()");
    }
    if (number < 2) {
      throw Exception(
          "Error: setQuickReconnectMaxRetries() - invalid argument; expected an integer higher or equal to 2");
    }
    configuration.setQuickReconnectMaxRetries(number);
    logger.info("Configuring quickReconnect max retries to: $number");
  }

  setQuickReconnectInitialDelay(int interval) {
    if (running) {
      throw Exception(
          "Error: setQuickReconnectInitialDelay() - already connected; use this method before connect()");
    }
    if (interval < 1) {
      throw Exception(
          "Error: setQuickReconnectInitialDelay() - invalid argument; expected an integer higher or equal to 1");
    }
    configuration.setQuickReconnectInitialDelay(interval * 1000);
    logger.info("Configuring quickReconnectInitial delay to: $interval");
  }

  setReconnectPolicy(String policy) {
    if (running) {
      throw Exception(
          "Error: setReconnectPolicy() - already connected; use this method before connect()");
    }
    if (policy != MigratoryDataClient.CONSTANT_WINDOW_BACKOFF &&
        policy != MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF) {
      throw Exception(
          "Error: setReconnectPolicy() - invalid argument; expected MigratoryDataClient.CONSTANT_WINDOW_BACKOFF or MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF");
    }
    configuration.setReconnectPolicy(policy);
    logger.info("Configuring reconnect policy to: $policy");
  }

  setReconnectTimeInterval(int interval) {
    if (running) {
      throw Exception(
          "Error: setReconnectTimeInterval() - already connected; use this method before connect()");
    }
    if (interval < 1) {
      throw Exception(
          "Error: setReconnectTimeInterval() - invalid argument; expected an integer higher or equal to 1");
    }
    configuration.setReconnectTimeInterval(interval * 1000);
    logger.info("Configuring setReconnectTime interval  to: $interval");
  }

  setReconnectMaxDelay(int delay) {
    if (running) {
      throw Exception(
          "Error: setReconnectMaxDelay() - already connected; use this method before connect()");
    }
    if (delay < 1) {
      throw Exception(
          "Error: setReconnectMaxDelay() - invalid argument; expected an integer higher or equal to 1");
    }
    configuration.setReconnectMaxDelay(delay * 1000);
    logger.info("Configuring setReconnectMax delay  to: $delay");
  }

  void setHttpHeader(String header, String value, bool preserveHeaderCase) {
    if (running) {
      throw Exception(
          "Error: setHttpHeader() - already connected; use this method before connect()");
    }
    configuration.setHttpHeader(header, value, preserveHeaderCase);
    logger.info(
        "Configuring http header `$header` to `$value` with preserveHeaderCase `$preserveHeaderCase`");
  }

  void setSocketTimeoutSeconds(int seconds) {
    if (running) {
      throw Exception(
          "Error: setSocketTimeoutSeconds() - already connected; use this method before connect()");
    }
    if (seconds < 1) {
      throw Exception(
          "Error: setSocketTimeoutSeconds() - invalid argument; expected an integer higher or equal to 1");
    }

    configuration.setSocketTimeoutSeconds(seconds);
    logger.info("Configuring socket timeout to `$seconds` seconds.");
  }

  void setExternalToken(String token) {
    if (running) {
      throw Exception(
          "Error: setExternalToken() - already connected; use this method before connect()");
    }
    configuration.setExternalToken(token);
    logger.info(
        "Configuring external token to `$token`");
  }
}
