import 'dart:typed_data';

import 'websocket/websocket_stub.dart'
    if (dart.library.html) 'websocket/websocket_html.dart'
    if (dart.library.io) 'websocket/websocket_io.dart';

import 'compression/compression_stub.dart'
    if (dart.library.html) 'compression/compression_html.dart'
    if (dart.library.io) 'compression/compression_io.dart';

import '../logger/logger.dart';
import '../migratorydata_client.dart';
import '../migratorydata_message.dart';
import '../protocol/buffer.dart';
import '../protocol/message.dart';
import '../protocol/push_decoder.dart';
import '../protocol/push_encoder.dart';
import '../protocol/push_protocol.dart'
    show Header, EntitlementStatus, Operation, PushError;
import '../protocol/push_protocol.dart' as push_protocol;
import '../protocol/push_protocol_transport_codec.dart';
import 'cluster_info.dart';
import 'cluster_member_info.dart';
import 'configuration.dart';
import 'listener_notifier.dart';
import 'scheduler.dart';
import 'subject.dart';
import 'subject_manager.dart';
import 'util.dart';

class Connection {
  final Logger logger;

  final PushEncoder PUSH_ENCODER = PushEncoder();
  final PushDecoder PUSH_DECODER = PushDecoder();
  final PushProtocolTransportCodec TRANSPORT_CODEC =
      PushProtocolTransportCodec.instance;

  ClusterInfo cluster;
  ListenerNotifier listenerNotifier;
  int serversDownCount = 0;

  String currentConnectionId = "";

  bool reconnected = false;

  int reconnectRetries = 0;
  int sessionId = -1;
  bool sessionReceived = false;

  final SubjectManager subjectManager = SubjectManager();

  bool isServerDown = false;

  AppState appState = AppState.RUNNING;

  final Scheduler scheduler;

  final Configuration configuration;

  final ClientState clientState;

  String clusterToken = "";
  int maxMessageSize = 9999999999999;

  final websocketHelper = WebsocketHelperImpl();
  final compressionHelper = CompressionHelperImpl();

  Connection(this.listenerNotifier, this.scheduler, this.cluster,
      this.configuration, this.clientState, this.logger);

  void connect() async {
    String connectionId = DateTime.now().millisecondsSinceEpoch.toString();
    setCurrentConnectionId(connectionId);

    disconnect();

    ClusterMemberInfo clusterMember = cluster.getNextMember();

    bool connected = await websocketHelper.connect(
        clusterMember, connectionId, this, logger, configuration, scheduler);

    if (connected) {
      scheduler.resetAndSetKeepAliveTimer(
          connectionId, this, KeepAlive.CONNECT);
      scheduler.setPingOperation(this);
      doConnect();
    }
  }

  void doConnect() {
    Buffer buf = Buffer(256);
    PUSH_ENCODER.encodeConnect(
        buf,
        configuration.getEntitlementToken(),
        configuration.getSessionType(),
        configuration.getUserAgent(),
        configuration.getVersion(),
        configuration.getExternalToken(),
        configuration.getMonotonicId());
    write(buf);
  }

  void armNotifyServerDown() {
    isServerDown = false;
    serversDownCount = 0;
  }

  void connectionEstablished() {
    if (logger.isDebugEnabled()) {
      logger.debug(
          "Connected to the cluster member: ${cluster.getCurrentMember()}");
    }
    armNotifyServerDown();
    listenerNotifier.offerStatus(MigratoryDataClient.NOTIFY_SERVER_UP,
        cluster.getCurrentMember().getUnparseAddress());
  }

  void connectionFailed(String disconnectInfo) {
    logger.error("[$disconnectInfo]  [${cluster.getCurrentMember()}]");
    if (logger.isDebugEnabled()) {
      logger.debug(
          "Lost connection with the cluster member: ${cluster.getCurrentMember()}");
    }
    if (!sessionReceived) {
      serversDownCount++;
      if (!isServerDown) {
        if (serversDownCount >= configuration.getServersDownBeforeNotify()) {
          isServerDown = true;
          listenerNotifier.offerStatus(MigratoryDataClient.NOTIFY_SERVER_DOWN,
              cluster.getCurrentMember().getUnparseAddress());
        }
      }
    }
  }

  void tryToReconnect() {
    disconnect();
    if (appState == AppState.TERMINATE) {
      return;
    }
    cluster.setMemberInactive(cluster.getCurrentMember());
    reconnected = true;
    connect();
  }

  void disconnect() {
    websocketHelper.close();
    scheduler.cancel();

    reset();
  }

  void dispose() {
    appState = AppState.TERMINATE;
    disconnect();
    scheduler.stop();
  }

  void subscribe(List<String> subjects, int history) {
    if (subjects.isEmpty) {
      return;
    }
    List<String> subscribeTo =
        Util.difference(subjects, subjectManager.getAllSubjects());
    if (subscribeTo.isEmpty) {
      return;
    }

    subjectManager.addSubjects(subscribeTo, history);

    if (clientState.getState() == State.CONNECTED) {
      subscribe0(subscribeTo);
    }
  }

  void subscribe0(List<String> subjects) {
    Buffer buf = Buffer(256);
    for (String s in subjects) {
      Subject? sObject = subjectManager.getSubject(s);
      if (sObject != null) {
        doSubscribe(buf, sObject);
      }
    }
    write(buf);
  }

  void doSubscribe(Buffer buf, Subject subject) {
    PUSH_ENCODER.encodeSubscribe(buf, subject, sessionId);
  }

  void unsubscribe(List<String> subjects) {
    if (subjects.isEmpty) {
      return;
    }
    List<String> unsubscribeFrom =
        Util.intersection(subjects, subjectManager.getAllSubjects());
    if (unsubscribeFrom.isEmpty) {
      return;
    }

    List<Subject> removedSubjects =
        subjectManager.removeSubjects(unsubscribeFrom);

    if (clientState.getState() == State.CONNECTED) {
      unsubscribe0(removedSubjects);
    }
  }

  void unsubscribe0(List<Subject> subjects) {
    Buffer buf = Buffer(256);
    for (Subject s in subjects) {
      PUSH_ENCODER.encodeUnsubscribe(buf, sessionId, s);
    }
    write(buf);
  }

  void channelRead(Uint8List buf, String connectionId) {
    List<Message> messages = PUSH_DECODER.parseWebsocketFrame(buf, logger);
    if (messages.isNotEmpty) {
      processMessages(messages, connectionId);
    } else {
      scheduler.resetAndSetKeepAliveTimer(connectionId, this, KeepAlive.RESET);
    }
  }

  void reset() {
    clientState.setState(State.DISCONNECTED);

    sessionId = -1;
    sessionReceived = false;
  }

  void write(Buffer msg) {
    msg.flip();

    Uint8List data = Uint8List(msg.length);
    List.copyRange(data, 0, msg.array(), 0, data.length);

    websocketHelper.write(data);
  }

  void processMessages(List<Message> messages, String connectionId) {
    for (Message msg in messages) {
      switch (msg.getOperation()) {
        case Operation.PUBLISH:
        case Operation.ENTITLEMENT_CHECK:
        case Operation.CLIENT_PUBLISH_RESPONSE:
        case Operation.RECOVERY_CACHE:
        case Operation.CONNECT:
        case Operation.SUBSCRIBE:
        case Operation.UNSUBSCRIBE:
        case Operation.PING:
        case Operation.STATUS_EVENT:
          onMessage(msg, connectionId);
          break;
        case Operation.NULL:
        default:
          logger.warn("No existing operation for message : $msg");
      }
    }
  }

  void publish(MigratoryDataMessage message) {
    if (clientState.getState() != State.CONNECTED) {
      failedToDeliverMessage(
          MigratoryDataClient.NOTIFY_PUBLISH_FAILED, message);
      return;
    }

    publish0(message);
  }

  void publish0(MigratoryDataMessage message) {
    String replySubject = message.getReplySubject();
    if (replySubject.isNotEmpty &&
        !subjectManager.containsSubject(replySubject)) {
      subscribe([replySubject], 0);
    }

    Buffer buf = Buffer(256);
    PUSH_ENCODER.encodePublish(buf, message, sessionId);

    if (buf.length > maxMessageSize) {
      failedToDeliverMessage(
          MigratoryDataClient.NOTIFY_MESSAGE_SIZE_LIMIT_EXCEEDED, message);
      return;
    }

    write(buf);
  }

  void failedToDeliverMessage(
      String notification, MigratoryDataMessage message) {
    if (message.getClosure().isNotEmpty) {
      listenerNotifier.offerStatus(notification, message.getClosure());
    }
  }

  void sendAuthorizationToken(String token) {
    Buffer buf = Buffer(256);
    PUSH_ENCODER.encodeAuthorizationTokenUpdate(buf, token, sessionId);
    write(buf);
  }

  void pause() {
    if (appState != AppState.RUNNING) {
      return;
    }
    if (logger.isDebugEnabled()) {
      logger.debug("Client enters in pause state.");
    }
    appState = AppState.PAUSE;
    disconnect();
  }

  void resume() {
    if (appState != AppState.PAUSE) {
      return;
    }
    if (logger.isDebugEnabled()) {
      logger.debug("Client enters in resume state.");
    }
    appState = AppState.RUNNING;
    armNotifyServerDown();
    tryToReconnect();
  }

  void ping() {
    Buffer buf = Buffer(256);
    PUSH_ENCODER.encodePing(buf, sessionId);
    write(buf);
  }

  String getCurrentConnectionId() {
    return currentConnectionId;
  }

  void setCurrentConnectionId(String uuid) {
    currentConnectionId = uuid;
  }

  AppState getAppState() {
    return appState;
  }

  int incrementAndGetReconnectRetries() {
    reconnectRetries++;
    return reconnectRetries;
  }

  void reconnect(String uuid, String disconnectReason, String from) {
    if (uuid == getCurrentConnectionId()) {
      clientState.setState(State.DISCONNECTED);
      disconnect();
      scheduler.reconnect(uuid, this, disconnectReason);
    }
  }

  int getReconnectRetries() {
    return reconnectRetries;
  }

  @override
  String toString() {
    return "Connection.class=$cluster.getCurrentMember()";
  }

  void onMessage(Message msg, String connectionId) {
    scheduler.resetAndSetKeepAliveTimer(connectionId, this, KeepAlive.RESET);

    Map<Header, Object> headers = msg.getHeaders();
    switch (msg.getOperation()) {
      case Operation.PUBLISH:
        processNewDataMessage(headers, connectionId);
        break;
      case Operation.SUBSCRIBE:
        processSubscribeResponse(headers);
        break;
      case Operation.CONNECT:
        processConnectResponse(headers, connectionId);
        break;
      case Operation.UNSUBSCRIBE:
        processUnsubscribeResponse();
        break;
      case Operation.ENTITLEMENT_CHECK:
        processEntitlementCheckResponse(headers);
        break;
      case Operation.CLIENT_PUBLISH_RESPONSE:
        processClientPublishResponse(headers);
        break;
      case Operation.RECOVERY_CACHE:
        processCacheRecoveryStatus(headers);
        break;
      case Operation.STATUS_EVENT:
        processStatusEvent(headers);
        break;
      default:
        logger.warn("No existing operation for message : $msg");
    }
  }

  void processConnectResponse(
      Map<Header, Object> headers, String connectionId) {
    int? sessionIdReceived = headers[Header.SESSION_ID] as int?;
    if (sessionIdReceived != null) {
      connectionEstablished();
      sessionId = sessionIdReceived;
      sessionReceived = true;
      reconnectRetries = 0;

      int? keepAliveTime = headers[Header.SERVER_CLIENT_PING_TIME] as int?;
      if (keepAliveTime != null) {
        scheduler.setKeepAliveTimeMillis(keepAliveTime);
        scheduler.resetAndSetKeepAliveTimer(
            connectionId, this, KeepAlive.RESET);
      }
      clientState.setState(State.CONNECTED);

      String? receivedClusterToken = headers[Header.CLUSTER_TOKEN] as String?;
      handleClusterToken(receivedClusterToken ?? "");

      int? receivedMaxMessageSize = headers[Header.MAX_MESSAGE_SIZE] as int?;
      if (receivedMaxMessageSize != null) {
        maxMessageSize = receivedMaxMessageSize;
      }

      String? reason = headers[Header.ENTITLEMENT_STATUS] as String?;
      int? error = headers[Header.ERROR] as int?;
      if (error != null && error == PushError.CONNECT_DENY.value) {
        listenerNotifier.offerStatus(MigratoryDataClient.NOTIFY_CONNECT_DENY, reason ?? "");
      } else {
        listenerNotifier.offerStatus(MigratoryDataClient.NOTIFY_CONNECT_OK, reason ?? "");
      }

      List<String> subjects = subjectManager.getAllSubjects();
      if (subjects.isNotEmpty) {
        subscribe0(subjects);
      }
    }
  }

  void processSubscribeResponse(Map<Header, Object> headers) {}

  void processUnsubscribeResponse() {}

  void processEntitlementCheckResponse(Map<Header, Object> headers) {
    String? entitlement = headers[Header.ENTITLEMENT_STATUS] as String?;
    String? subject = headers[Header.SUBJECT] as String?;
    if (entitlement != null && subject != null) {
      bool subscribeFailed = true;
      String subscribeCheckStatus = MigratoryDataClient.NOTIFY_SUBSCRIBE_DENY;
      if (entitlement == EntitlementStatus.ALLOW.value) {
        subscribeCheckStatus = MigratoryDataClient.NOTIFY_SUBSCRIBE_ALLOW;
        subscribeFailed = false;
      } else if (entitlement == EntitlementStatus.DENY.value) {
        subscribeCheckStatus = MigratoryDataClient.NOTIFY_SUBSCRIBE_DENY;
      }
      if (subscribeFailed) {
        subjectManager.removeSubjects([subject]);
      }
      listenerNotifier.offerStatus(subscribeCheckStatus, subject);
    }
  }

  void processNewDataMessage(Map<Header, Object> headers, String connectionId) {
    String? subject = headers[Header.SUBJECT] as String?;

    if (subject == null) {
      return;
    }
    Subject? sbj = subjectManager.getSubject(subject);
    if (sbj == null) {
      return;
    }

    String? receivedClusterToken = headers[Header.CLUSTER_TOKEN] as String?;
    handleClusterToken(receivedClusterToken ?? "");

    Uint8List? data = headers[Header.DATA] as Uint8List?;
    if (data == null) {
      return;
    }
    String? closure = headers[Header.CLOSURE] as String?;

    bool retained = false;
    int? retainedValue = headers[Header.PUBLICATION_RETAINED] as int?;
    if (retainedValue != null && retainedValue == 1) {
      retained = true;
    }

    bool compression = false;
    int? compressionValue = headers[Header.COMPRESSION] as int?;
    if (compressionValue != null && compressionValue == 1) {
      compression = true;
    }

    if (compression) {
      data = compressionHelper.base64DecodeAndzLibUncompress(data);
    }

    MessageType mdsMessageType = MessageType.UPDATE;
    String? messageType = headers[Header.MESSAGE_TYPE] as String?;
    if (messageType != null) {
      if (messageType == push_protocol.MessageType.SNAPSHOT.value) {
        mdsMessageType = MessageType.SNAPSHOT;
      } else if (messageType == push_protocol.MessageType.RECOVERED.value) {
        mdsMessageType = MessageType.RECOVERED;
      } else if (messageType == push_protocol.MessageType.HISTORICAL.value) {
        mdsMessageType = MessageType.HISTORICAL;
      }
    }

    QoS qos = QoS.GUARANTEED;
    int? qosValue = headers[Header.PUBLICATION_QOS] as int?;
    if (qosValue != null && qosValue == QoS.STANDARD.index) {
      qos = QoS.STANDARD;
    }

    String? replySubject = headers[Header.REPLY_SUBJECT] as String?;

    MigratoryDataMessage message = MigratoryDataMessage(subject, data,
        closure: closure ?? "",
        qos: qos,
        retained: retained,
        replySubject: replySubject ?? "",
        compression: compression);
    message.messageType = mdsMessageType;

    if (qos == QoS.GUARANTEED) {
      int? seq = headers[Header.SEQ] as int?;
      int? seqId = headers[Header.SEQ_ID] as int?;

      if (seq != null && seqId != null) {
        message.seq = seq;
        message.epoch = seqId;

        OrderGuaranteed order =
            Util.verifyUpdateOrder(sbj, seq, seqId, listenerNotifier, logger);
        if (order == OrderGuaranteed.SEQ_IN_ORDER) {
          listenerNotifier.offerMessage(message);
        } else if (order == OrderGuaranteed.SEQ_HIGHER) {
          reconnect(connectionId, Util.CONNECTION_ACTIVE_CLOSE_SEQ_HIGHER,
              "seq_higher");
        }
      }
    } else {
      listenerNotifier.offerMessage(message);
    }
  }

  void processClientPublishResponse(Map<Header, Object> headers) {
    if (headers.isEmpty) {
      return;
    }

    String? closure = headers[Header.CLOSURE] as String?;
    String? response = headers[Header.ENTITLEMENT_STATUS] as String?;
    if (closure != null && response != null) {
      String status = MigratoryDataClient.NOTIFY_PUBLISH_FAILED;
      if (response == Util.PUBLISH_REASON_DENY) {
        status = MigratoryDataClient.NOTIFY_PUBLISH_DENIED;
      } else if (response == Util.PUBLISH_REASON_OK) {
        status = MigratoryDataClient.NOTIFY_PUBLISH_OK;
      }
      listenerNotifier.offerStatus(status, closure);
    }
  }

  void processCacheRecoveryStatus(Map<Header, Object> headers) {
    String? subject = headers[Header.SUBJECT] as String?;
    String? status = headers[Header.MESSAGE_TYPE] as String?;

    if (subject == null || status == null) {
      logger.info("Invalid recovery status!");
      return;
    }

    if (logger.isDebugEnabled()) {
      logger.debug("Recovery status for subject: $subject is: $status");
    }

    if (Util.RECOVERY_CACHE_END == status) {
      List<String> subjects = subjectManager.getAllSubjects();
      for (String s in subjects) {
        Subject? sbj = subjectManager.getSubject(s);
        if (sbj != null) {
          String sbjRecoveryCacheStatus = sbj.getCacheRecoveryStatus();
          if (Util.RECOVERY_CACHE_OK == sbjRecoveryCacheStatus ||
              Util.RECOVERY_CACHE_OK_NEW_EPOCH == sbjRecoveryCacheStatus ||
              Util.RECOVERY_CACHE_OK_NO_NEW_MESSAGE == sbjRecoveryCacheStatus) {
            sbj.unmarkFromRecovery();
          } else {
            sbj.markForRecovery();
          }
        }
      }
    } else {
      Subject? sbj = subjectManager.getSubject(subject);
      if (sbj != null) {
        sbj.setCacheRecoveryStatus(status);
      }
    }
  }

  void handleClusterToken(String receivedClusterToken) {
    if (receivedClusterToken.isNotEmpty) {
      if (clusterToken.isEmpty) {
        clusterToken = receivedClusterToken;
      } else {
        if (!(receivedClusterToken == clusterToken)) {
          clusterToken = receivedClusterToken;
          // reset epoch and seq
          subjectManager.resetSubjects();
        }
      }
    }
  }

  void processStatusEvent(Map<Header, Object> headers) {
    String? status = headers[Header.ENTITLEMENT_STATUS] as String?;
    String? info = headers[Header.ENTITLEMENT_TOKEN] as String?;

    if (status != null && info != null) {
      listenerNotifier.offerStatus(status, info);
    }
  }
}

class ClientState {
  State state = State.DISCONNECTED;

  void setState(State state) {
    this.state = state;
  }

  State getState() {
    return state;
  }
}

enum State { DISCONNECTED, CONNECTED }

enum AppState { TERMINATE, PAUSE, RUNNING }

enum KeepAlive { CONNECT, RESET }
