import 'dart:async';
import 'dart:math';

import '../migratorydata_client.dart';
import 'configuration.dart';
import 'connection.dart';
import 'util.dart';

class Scheduler {
  static var DEFAULT_KEEP_ALIVE_TIMEOUT_MILLIS =
      30000 + (Random().nextDouble() + 10000).floor().round();
  static var PING_INTERVAL = 900000; // 15 minutes
  static var OPERATION_TIMEOUT_INTERVAL = 10000;

  int keepAliveTimeoutMillis = DEFAULT_KEEP_ALIVE_TIMEOUT_MILLIS;

  Configuration configuration;
  ClientState clientState;

  Timer? keepAliveTimer;
  Timer? pingOperationTimer;
  Timer? reconnectTimer;

  Scheduler(this.configuration, this.clientState);

  void resetAndSetKeepAliveTimer(
      String connectionId, Connection connection, KeepAlive keepAlive) {
    if (keepAliveTimer != null) {
      keepAliveTimer?.cancel();
    }

    int timeout = keepAliveTimeoutMillis;

    if (keepAlive == KeepAlive.CONNECT) {
      int retries = connection.getReconnectRetries();
      timeout = getNextTimeoutTime(retries, true);
    }

    if (timeout > 0) {
      setKeepAliveTimer(connectionId, connection, timeout);
    }
  }

  void reconnect(
      String connectionId, Connection connection, String disconnectInfo) {
    if (connection.getAppState() != AppState.RUNNING) {
      return;
    }
    String currentConnectionId = connection.getCurrentConnectionId();
    if (currentConnectionId.isEmpty || currentConnectionId != connectionId) {
      return;
    }

    connection.setCurrentConnectionId("");
    connection.connectionFailed(disconnectInfo);

    int retries = connection.incrementAndGetReconnectRetries();
    int timeout = getNextTimeoutTime(retries, false);
    setReconnectTimer(timeout, connection, disconnectInfo);
  }

  void setReconnectTimer(
      int time, final Connection connection, final String disconnectInfo) {
    if (reconnectTimer != null) {
      reconnectTimer?.cancel();
    }
    reconnectTimer = Timer(Duration(milliseconds: time), () {
      connection.tryToReconnect();
    });
  }

  void setKeepAliveTimeMillis(int value) {
    int interval = ((value * 1000) * 1.4).toInt();
    keepAliveTimeoutMillis = interval;
  }

  void setPingOperation(final Connection connection) {
    if (pingOperationTimer != null) {
      pingOperationTimer?.cancel();
    }
    pingOperationTimer = Timer(Duration(milliseconds: PING_INTERVAL), () {
      connection.ping();
      setPingOperation(connection);
    });
  }

  void cancel() {
    if (keepAliveTimer != null) {
      keepAliveTimer?.cancel();
    }
    if (pingOperationTimer != null) {
      pingOperationTimer?.cancel();
    }
    if (reconnectTimer != null) {
      reconnectTimer?.cancel();
    }
  }

  void stop() {
    cancel();
  }

  int getNextTimeoutTime(int retries, bool verifyTimeout) {
    int timeout = DEFAULT_KEEP_ALIVE_TIMEOUT_MILLIS;

    if (retries > 0) {
      Random rand = Random.secure();
      if (retries <= configuration.getQuickReconnectMaxRetries()) {
        timeout = ((retries * configuration.getQuickReconnectInitialDelay()) -
            ((rand.nextDouble() * configuration.getQuickReconnectInitialDelay())
                .floor()));
      } else {
        if (configuration.getReconnectPolicy() ==
            MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF) {
          int count = retries - configuration.getQuickReconnectMaxRetries();
          timeout = min(
              (configuration.getReconnectTimeInterval() * pow(2, count))
                      .toInt() -
                  (rand.nextDouble() *
                          configuration.getReconnectTimeInterval() *
                          count)
                      .floor(),
              configuration.getReconnectMaxDelay());
        } else {
          timeout = configuration.getReconnectTimeInterval();
        }
      }
      if (verifyTimeout && timeout < OPERATION_TIMEOUT_INTERVAL) {
        timeout = OPERATION_TIMEOUT_INTERVAL;
      }
    }
    return timeout;
  }

  void setKeepAliveTimer(String uuid, Connection connection, int time) {
    keepAliveTimer = Timer(Duration(milliseconds: time), () {
      String currentConnectionId = connection.getCurrentConnectionId();
      if (currentConnectionId.isEmpty || currentConnectionId != uuid) {
        return;
      }
      clientState.setState(State.DISCONNECTED);

      connection.disconnect();
      reconnect(uuid, connection, Util.CONNECTION_ACTIVE_CLOSE_KEEP_ALIVE);
    });
  }
}
