
import '../logger/logger.dart';
import '../migratorydata_client.dart';
import 'listener_notifier.dart';
import 'subject.dart';

class Util {
  static final pushSubjectRegex = RegExp(r'^\/([^\/]+\/)*([^\/]+|\*)$');

  static final String PUBLISH_REASON_OK = "OK";
  static final String PUBLISH_REASON_FAILED = "FAILED";
  static final String PUBLISH_REASON_DENY = "DENY";

  static final String CONNECTION_PASSIVE_CLOSE = "connection_passive_close";
  static final String CONNECTION_ACTIVE_CLOSE_KEEP_ALIVE =
      "connection_active_close_keep_alive"; // connection closed by the client after keepAliveTimeout
  static final String CONNECTION_ACTIVE_CLOSE_SEQ_HIGHER =
      "connection_active_close_seq_higher"; // connection closed by the client because the message seq is higher than the current seq
  static final String CONNECTION_ERROR_CLOSE = "connection_error";

  static final String RECOVERY_CACHE_OK = "cache_ok";
  static final String RECOVERY_CACHE_OK_NO_NEW_MESSAGE =
      "cache_ok_no_new_message";
  static final String RECOVERY_CACHE_OK_NEW_EPOCH = "cache_ok_new_epoch";
  static final String RECOVERY_CACHE_NO_CACHE = "no_cache";
  static final String RECOVERY_CACHE_NO_SEQ = "no_seq";
  static final String RECOVERY_CACHE_END = "end";

  static final String RECOVERY_CACHE_SUBJECT_EMPTY = "empty";

  static bool isPushSubjectValid(String subject) {
    return pushSubjectRegex.hasMatch(subject);
  }

  static bool areSubjectsValid(List<String> subjects) {
    for (String subject in subjects) {
      if (subject.isEmpty || !pushSubjectRegex.hasMatch(subject)) {
        return false;
      }
    }
    return true;
  }

  static List<String> difference(List<String> A, List<String> B) {
    List<String> difference = [];
    for (String a in A) {
      if (!B.contains(a)) {
        difference.add(a);
      }
    }
    return difference;
  }

  static List<String> intersection(List<String> A, List<String> B) {
    List<String> intersection = [];
    for (String a in A) {
      if (B.contains(a)) {
        intersection.add(a);
      }
    }
    return intersection;
  }

  static OrderGuaranteed verifyUpdateOrder(Subject sbj, int recvSeq,
      int recvSeqId, ListenerNotifier listener, Logger logger) {
    // different epoch, reset and continue.
    if (sbj.getSeqId() != recvSeqId) {
      sbj.setSeq(recvSeq);
      sbj.setSeqId(recvSeqId);
      return OrderGuaranteed.SEQ_IN_ORDER;
    }

    // if received seq is equal or smaller than the local seq then the message is ignored
    if (recvSeq <= sbj.getSeq()) {
      return OrderGuaranteed.SEQ_LOWER;
    }

    // if received seq is +1 than the local seq then the message is processed
    if (recvSeq == (sbj.getSeq() + 1)) {
      if (sbj.getCurrentSubscribeType() == SubscribeType.RECOVERY) {
        sbj.setCurrentSubscribeTypeToNormal();
        listener.offerStatus(
            MigratoryDataClient.NOTIFY_DATA_SYNC, sbj.getSubject());
      }

      sbj.setSeq(sbj.getSeq() + 1);
      return OrderGuaranteed.SEQ_IN_ORDER;
    }

    // there is a hole in the order of the messages
    // if there is a missing message when the session is active, then we disconnect the client and make failover.
    if (sbj.getMessagesReceivedUntilRecovery() > 0) {
      if (logger.isDebugEnabled()) {
        logger.debug(
            "Missing messages: expected message with sequence number: '${(sbj.getSeq() + 1)} ', received instead message with sequence number: $recvSeq !");
      }
      return OrderGuaranteed.SEQ_HIGHER;
    }

    if (logger.isDebugEnabled()) {
      logger.debug(
          "Reset sequence: '${(sbj.getSeq() + 1)} '. The new sequence is: '$recvSeq' !");
    }
    sbj.setSeq(recvSeq);
    listener.offerStatus(
        MigratoryDataClient.NOTIFY_DATA_RESYNC, sbj.getSubject());
    return OrderGuaranteed.SEQ_IN_ORDER;
  }
}

enum OrderGuaranteed { SEQ_IN_ORDER, SEQ_LOWER, SEQ_HIGHER }
