import 'dart:convert';
import 'dart:io';
import 'dart:math';
import 'dart:typed_data';

import '../../logger/logger.dart';
import '../cluster_member_info.dart';
import '../configuration.dart';
import '../connection.dart';
import '../scheduler.dart';
import '../util.dart';
import 'websocket_interface.dart';

class WebsocketHelperImpl extends WebsocketInterface {
  WebSocket? channel;

  Future<bool> connect(
      ClusterMemberInfo clusterMember,
      String connectionId,
      Connection connection,
      Logger logger,
      Configuration configuration,
      Scheduler scheduler) async {
    try {
      if (logger.isDebugEnabled()) {
        logger.debug("[io] Connecting to the cluster member: $clusterMember");
      }

      Random r = Random();
      String key = base64.encode(List<int>.generate(8, (_) => r.nextInt(255)));

      final client = HttpClient();

      if (configuration.getSocketTimeoutSeconds() > 0) {
        client.connectionTimeout =
            Duration(seconds: configuration.getSocketTimeoutSeconds());
      }
      
      final request =
          await client.openUrl('GET', Uri.parse(clusterMember.getUri()));

      request.headers
        ..set('Connection', 'Upgrade', preserveHeaderCase: true)
        ..set('Upgrade', 'websocket', preserveHeaderCase: true)
        ..set('Sec-WebSocket-Key', key, preserveHeaderCase: true)
        ..set('Sec-WebSocket-Protocol', 'pushv1', preserveHeaderCase: true)
        ..set('Host', '${clusterMember.host}:${clusterMember.port}',
            preserveHeaderCase: true)
        ..set('Origin', '${clusterMember.getOrigin()}',
            preserveHeaderCase: true)
        ..set('User-Agent', utf8.decode(configuration.getUserAgent()),
            preserveHeaderCase: true)
        ..set('Sec-WebSocket-Version', '13', preserveHeaderCase: true);

      Map<String, HttpHeaders> httpHeaders = configuration.getHttpHeaders();
      httpHeaders.forEach((key, value) {
        request.headers.set(value.header, value.value,
            preserveHeaderCase: value.preserveHeaderCase);
      });

      final response = await request.close();

      if (response.statusCode != 101) {
        throw Exception("Invalid status code, failed to connect!");
      }

      final socket = await response.detachSocket();

      channel = WebSocket.fromUpgradedSocket(socket, serverSide: false);

      channel?.handleError((err, stack) => { logger.info("onError channel websocket:${err.toString()}") });

      channel?.listen(
        (data) {
          if (data is String) {
            connection.channelRead(
                Uint8List.fromList(utf8.encode(data)), connectionId);
          } else {
            connection.channelRead(data, connectionId);
          }
        },
        onDone: () {
          logger.info("onDone websocket.");
          connection.reconnect(
              connectionId, Util.CONNECTION_PASSIVE_CLOSE, "onDone");
        },
        onError: (err, stack) {
          logger.info("onError websocket:${err.toString()}");
          connection.reconnect(
              connectionId, Util.CONNECTION_PASSIVE_CLOSE, "onError");
        },
        cancelOnError: true,
      );
    } catch (e) {
      if (logger.isInfoEnabled()) {
        logger.info("Failed to connect: $clusterMember, message: $e");
      }

      scheduler.reconnect(
          connectionId, connection, Util.CONNECTION_ERROR_CLOSE);
      return false; // failed to create socket connection
    }
    return true;
  }

  @override
  void close() {
    channel?.close();
    channel = null;
  }

  @override
  void write(Uint8List data) {
    if (channel?.readyState == WebSocket.open) {
      channel?.add(data);
    } else {
      print("[io] channel ready state != open");
    }
  }
}
