import 'core/client.dart';
import 'utils/monotonic_id.dart';
import 'migratorydata_listener.dart';
import 'migratorydata_log_listener.dart';
import 'migratorydata_message.dart';

/// This class implements all the necessary operations for connecting to a cluster of one or more MigratoryData servers,
/// subscribing to one or more subjects, getting real-time messages for the subscribed subjects, and publishing messages.
class MigratoryDataClient {
  late ClientImpl clientImpl;

  /// A constant which indicates that the client successfully connected to a MigratoryData server.
  static const String NOTIFY_SERVER_UP = "NOTIFY_SERVER_UP";

  /// A constant which indicates that the client failed to connect to a MigratoryData server.
  static const String NOTIFY_SERVER_DOWN = "NOTIFY_SERVER_DOWN";

  /// A constant which indicates that after a fail-over reconnection, the client successfully synchronized a subscribed
  /// subject with the latest message available for that subject, as well as with all messages published during the
  /// fail-over period for that subject.
  static const String NOTIFY_DATA_SYNC = "NOTIFY_DATA_SYNC";

  /// A constant which indicates that after a fail-over reconnection, the client successfully synchronized a subscribed
  /// subject with the latest message available for that subject, but not with the potential messages published during
  /// the fail-over, therefore behaving as a fresh client.
  static const String NOTIFY_DATA_RESYNC = "NOTIFY_DATA_RESYNC";

  /// A constant which indicates that the client was authorized to subscribe to a subject.
  static const String NOTIFY_SUBSCRIBE_ALLOW = "NOTIFY_SUBSCRIBE_ALLOW";

  /// A constant which indicates that the client was not authorized to subscribe to a subject.
  static const String NOTIFY_SUBSCRIBE_DENY = "NOTIFY_SUBSCRIBE_DENY";

  /// A constant which indicates that the client successfully published a message.
  static const String NOTIFY_PUBLISH_OK = "NOTIFY_PUBLISH_OK";

  /// A constant which indicates that the client was unable to publish a message.
  static const String NOTIFY_PUBLISH_FAILED = "NOTIFY_PUBLISH_FAILED";

  /// A constant which indicates that the client was unable to publish a message because the size of the message
  /// exceeds the message size limit allowed by the server (see the server parameter \c MAX_MESSAGE_SIZE).
  static const String NOTIFY_MESSAGE_SIZE_LIMIT_EXCEEDED =
      "NOTIFY_MESSAGE_SIZE_LIMIT_EXCEEDED";

  /// A constant which indicates that the client was not authorized to publish a message.
  static const String NOTIFY_PUBLISH_DENIED = "NOTIFY_PUBLISH_DENIED";

  /// A constant which indicates that the client was authorized to connect using the entitlement token defined on the client side.
  ///
  /// NOTE: This notification applies when using a `Custom` authorization extension built with the
  /// MigratoryData Authorization Extension API version 2 or later. For the entitlement methods `None`, `Basic`,
  /// or `Custom` authorization extension built with the previous version of the MigratoryData Authorization Extension API,
  /// this notification is always sent no matter the entitlement token is valid or not, the verification of the entitlement
  /// token being made only during subscribe and publish operations.
  static const String NOTIFY_CONNECT_OK = "NOTIFY_CONNECT_OK";


  /// A constant which indicates that the client was denied to connect using the entitlement token defined on the client side.
  ///
  /// NOTE: This notification applies when using a `Custom` authorization extension built with the
  /// MigratoryData Authorization Extension API version 2 or later. For the entitlement methods `None`, `Basic`,
  /// or `Custom` authorization extension built with the previous version of the MigratoryData Authorization Extension API,
  /// this notification is never sent no matter the entitlement token is valid or not, the verification of the entitlement
  /// token being made only during subscribe and publish operations.
  static const String NOTIFY_CONNECT_DENY = "NOTIFY_CONNECT_DENY";

  /// A constant used to define the reconnect policy.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink for more details about this policy.
  static const String CONSTANT_WINDOW_BACKOFF = "CONSTANT_WINDOW_BACKOFF";

  /// A constant used to define the reconnect policy.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink for more details about this policy.
  static const String TRUNCATED_EXPONENTIAL_BACKOFF =
      "TRUNCATED_EXPONENTIAL_BACKOFF";

  /// Create a MigratoryDataClient object.
  MigratoryDataClient() {
    clientImpl = ClientImpl();
  }

  /// Attach a listener for handling log messages outputted by the library.
  ///
  /// It is advisable to configure this listener first to log as much as possible. If no log listener is set then, by
  /// default the client will log to the console.
  ///
  /// \param logListener   an implementation of the \link MigratoryDataLogListener \endlink interface
  /// \param logLevel      a particular \link MigratoryDataLogLevel \endlink configured as the logging threshold
  void setLogListener(
      MigratoryDataLogListener logListener, MigratoryDataLogLevel logLevel) {
    clientImpl.setLogListener(logListener, logLevel);
  }

  /// Attach a listener for handling the received real-time messages as well as the status notifications.
  ///
  /// \param listener   an implementation of the \link MigratoryDataListener \endlink interface
  setListener(MigratoryDataListener listener) {
    clientImpl.setListener(listener);
  }

  /// Get the listener for handling real-time messages and status notifications.
  ///
  /// This is the listener set with \link MigratoryDataClient.setListener() \endlink.
  ///
  /// \return   the listener for handling real-time messages and status notifications
  MigratoryDataListener getListener() {
    return clientImpl.getListener();
  }

  /// Specify a cluster of one or more MigratoryData servers to which the client will connect to.
  ///
  /// For example, to connect to a cluster formed of two MigratoryData servers installed at the addresses
  /// \c p1.example.com and \c p2.example.com, and configured to accept clients on the standard HTTP port \c 80,
  /// the following code can be used:
  ///
  /// <pre>
  ///    client.setServers(new String[] {"p1.example.com:80", "p2.example.com:80"});
  /// </pre>
  ///
  /// To achieve load-balancing, the library connects the client to a MigratoryData server chosen randomly from the
  /// \c servers list. In this way, the load is balanced among all the members of the cluster.
  ///
  /// <blockquote>
  /// <small>
  /// Moreover, the library supports weighted load-balancing. This feature is especially useful if the MigratoryData
  /// servers in the cluster are installed on machines with different capacities. You can assign to each member of the
  /// cluster a \em weight ranging from \c 0 to \c 100. This weight assignment is a hint provided to the library to
  /// select with a higher probability a MigratoryData server with a higher weight either initially when the client
  /// connects to the cluster or later during a fail-over reconnection.
  ///
  /// Supposing the address \c p1.example.com corresponds to a machine that is twice more powerful than the machine
  /// having the address \c p2.example.com, then you can assign to \c p1.example.com a weight \c 100 and to
  /// \c p2.example.com a weight \c 50 by prefixing each address with the assigned weight as follows:
  ///
  /// <pre>
  ///     client.setServers(new String[] {"100 p1.example.com:80", "50 p2.example.com:80"});
  /// </pre>
  ///
  /// The library assigns a default weight \c 100 to the addresses not prefixed with a specific weight.
  /// </small>
  /// </blockquote>
  ///
  /// To achieve failover, if the connection between the client and a MigratoryData server is broken, then the library
  /// will automatically detect the failure and will select another MigratoryData server from the \c servers list. If
  /// the client fails to connect to the newly selected server, a status notification
  /// MigratoryDataClient.NOTIFY_SERVER_DOWN will be triggered (unless you modify the number of failed connection
  /// attempts with \link MigratoryDataClient.notifyAfterFailedConnectionAttempts() \endlink), and a new MigratoryData
  /// server of the cluster will be selected again and again until the client will be able to connect to one of the
  /// MigratoryData servers of the cluster. When successfully connected the library will trigger a status notification
  /// MigratoryDataClient.NOTIFY_SERVER_UP.
  ///
  /// \param servers   an array of strings where each string represents the network address (IP address or DNS domain
  ///                  name and its corresponding port) of a MigratoryData server, optionally prefixed by a weight
  ///                  ranging from \c 0 to \c 100; if the weight prefix is not provided to an address, then the \c 100
  ///                  weight will be automatically assigned to that address
  setServers(List<String> servers) {
    clientImpl.setServers(servers);
  }

  ///  Connect to the MigratoryData cluster.
  ///
  ///  Use this method to connect to one of the MigratoryData servers specified with \link MigratoryDataClient.setServers() \endlink.
  void connect() {
    clientImpl.connect();
  }

  /// Subscribe to one or more subjects.
  ///
  ///  As an example, supposing messages are market data updates having as subjects stock names. Then, to subscribe for
  ///  the messages having as subjects \c /stocks/NYSE/IBM and \c /stocks/Nasdaq/MSFT use:
  ///
  /// <pre>
  /// 	List<String> subjects = new ArrayList<String>();
  /// 	subjects.add("/stocks/NYSE/IBM");
  /// 	subjects.add("/stocks/Nasdaq/MSFT");
  /// 	client.subscribe(subjects);
  /// </pre>
  ///
  ///  The subjects are strings having a particular syntax. See the "Concepts" section of the <em>Architecture Guide</em>
  ///  to learn the syntax of the subjects.
  ///
  /// \param subjects   subjects to subscribe
  subscribe(List<String> subjects) {
    clientImpl.subscribe(subjects, 0);
  }

  /// Get the historical messages for one or more subjects, if any, then subscribe to those subjects.
  ///
  ///  Attempt to get the number of historical messages as defined by the argument \c numberOfHistoricalMessages,
  ///  for each subject in the argument \c subjects, then subscribe to than subject.
  ///
  ///  When Guaranteed Message Delivery is enabled, each MigratoryData server of the cluster maintains an in-memory
  ///  cache with historical messages for each subject. The cache of each subject is available in all servers of the
  ///  cluster. The maximum number of messages held in cache is defined by the parameter \c MaxCachedMessagesPerSubject
  ///  of the MigratoryData server which defaults to 1,000 messages. The historical messages are continuously removed
  ///  from the cache, but it is guaranteed that they are available in the cache at least the number of seconds defined
  ///  by the parameter \c CacheExpireTime which defaults to \c 180 seconds.
  ///
  ///  <p>
  ///      Note: If the value of \c numberOfHistoricalMessages is higher than the number of historical messages
  ///            available in the cache, then the client will receive only the messages available in the cache. As a
  ///            consequence, if you use a value higher than the value of the value of the parameter
  ///            \c MaxCachedMessagesPerSubject of the MigratoryData server (which defaults to 1000), then you will get
  ///            the entire cache before subscribing to the subjects specified with this API method.
  ///
  ///            If the value of \c numberOfHistoricalMessages is \c 0, then this API method is equivalent to the API
  ///            method  \link MigratoryDataClient.subscribe() \endlink.
  ///  </p>
  ///
  /// \param subjects                     subjects to subscribe
  /// \param numberOfHistoricalMessages   the number of historical messages to be retrieved from the cache of the server
  subscribeWithHistory(List<String> subjects, int numberOfHistoricalMessages) {
    clientImpl.subscribe(subjects, numberOfHistoricalMessages);
  }

  /// Unsubscribe from one or more subjects.
  ///
  /// \param subjects   subjects to unsubscribe
  unsubscribe(List<String> subjects) {
    clientImpl.unsubscribe(subjects);
  }

  /// Configure whether or not to use an encrypted SSL/TLS connection to communicate with the server.
  ///
  /// When using encryption you should connect to the ports of the MigratoryData server that are configured with the
  /// parameter \c ListenEncrypted to listen for encrypted connections.
  /// \param encrypted   indicate whether or not to use an encrypted SSL/TLS connection to communicate with the server
  setEncryption(bool encrypted) {
    clientImpl.setEncryption(encrypted);
  }

  /// Assign an entitlement token to the client or update its entitlement token.
  ///
  /// To define which users of your application are entitled to subscribe to which subjects and/or to publish messages
  /// for which subjects, you will first have to configure an entitlement method using the parameter \c Entitlement of
  /// the MigratoryData server. Please refer to the documentation of the \c Entitlement parameter in the
  /// `Configuration Guide` to learn how to configure an entitlement method.
  ///
  /// \param token   an entitlement token.
  setEntitlementToken(String token) {
    clientImpl.setEntitlementToken(token);
  }

  /// \brief Return the list of subscribed subjects.
  ///
  /// \return   the subscribed subjects
  List<String> getSubjects() {
    return clientImpl.getSubjects();
  }

  ///  Define the number of failed attempts to connect to one or more MigratoryData servers before triggering a status
  ///  notification MigratoryDataClient.NOTIFY_SERVER_DOWN.
  ///
  /// \param retries   the number of the failed attempts to connect to one or more MigratoryData servers before
  ///                  triggering a status notification MigratoryDataClient.NOTIFY_SERVER_DOWN; the default is \c 1
  notifyAfterFailedConnectionAttempts(int retries) {
    clientImpl.notifyAfterFailedConnectionAttempts(retries);
  }

  /// Disconnect from the connected MigratoryData server and dispose the resources used by the connection.
  ///
  /// This method should be called when the connection is no longer necessary.
  disconnect() {
    clientImpl.disconnect();
  }

  ///  Publish a message.
  ///
  ///  If the message includes a closure, then a status notification will be provided via
  ///  \link MigratoryDataListener.onStatus() \endlink to inform whether the message publication has been successful
  ///  or not.
  ///
  /// \param message   a MigratoryDataMessage message
  publish(MigratoryDataMessage message) {
    clientImpl.publish(message);
  }

  /// Define the maximum number of retries for the Quick Reconnect fail-over phase.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink to learn about the Quick Reconnect phase.
  ///
  /// \param retries   the maximum number of quick reconnect retries; the default is \c 3
  setQuickReconnectMaxRetries(int retries) {
    clientImpl.setQuickReconnectMaxRetries(retries);
  }

  /// \brief Define the number of seconds to wait before attempting to reconnect to the cluster after a connection
  /// failure is detected.
  ///
  /// <p>
  /// Reconnection Phases and Policies
  /// </p>
  ///
  /// When a connection failure is detected, the API will attempt to reconnect to the servers of the MigratoryData
  /// cluster as follows: First, it will attempt to reconnect up to a number of times as defined by
  /// \link MigratoryDataClient.setQuickReconnectMaxRetries() \endlink using small delays between retries ("Quick
  /// Reconnection Phase"). If the connection cannot be established after the Quick Reconnection Phase, then the API
  /// will attempt to reconnect less frequently according to the policy defined by
  /// \link MigratoryDataClient.setReconnectPolicy() \endlink.
  ///
  /// The delays between retries are computed according to the following algorithm where the values of the variables
  /// involved are defined by the API methods having substantially the same names:
  ///
  /// <pre>
  ///    Quick Reconnect Phase (retries <= quickReconnectMaxRetries)
  ///    -----------------------------------------------------------
  ///
  ///       (retries starts with 1 and increment by 1 at each quick reconnect)
  ///
  ///       reconnectDelay = quickReconnectInitialDelay * retries - random(0, quickReconnectInitialDelay)
  ///
  ///    After Quick Reconnect Phase (retries > quickReconnectMaxRetries)
  ///    ----------------------------------------------------------------
  ///
  ///       (reset retries to start with 1 and increment by 1 at each reconnect)
  ///
  ///       If reconnectPolicy is CONSTANT_WINDOW_BACKOFF, then
  ///
  ///          reconnectDelay = reconnectTimeInterval
  ///
  ///       else if reconnectPolicy is TRUNCATED_EXPONENTIAL_BACKOFF, then
  ///
  ///          reconnectDelay = min(reconnectTimeInterval * (2 ^ retries) - random(0, reconnectTimeInterval * retries), reconnectMaxDelay)
  /// </pre>
  ///
  /// For a few users which are subject to temporary, atypical network conditions, if `reconnectDelay` computed
  /// with the algorithm above is less than 10 seconds, then it is rounded to 10 seconds.
  ///
  /// \param seconds   the number of seconds to wait before attempting to reconnect to the cluster; the default is \c 5
  setQuickReconnectInitialDelay(int seconds) {
    clientImpl.setQuickReconnectInitialDelay(seconds);
  }

  /// Define the reconnect policy to be used after the Quick Reconnect phase.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink to learn about the Quick Reconnect phase
  /// and the reconnect schedule for the policy defined by this method.
  ///
  /// \param  policy   the reconnect policy to be used after the Quick Reconnect phase; the possible values are:
  ///                  \link MigratoryDataClient.CONSTANT_WINDOW_BACKOFF \endlink and
  ///                  \link MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF \endlink; the default is the last one
  setReconnectPolicy(String policy) {
    clientImpl.setReconnectPolicy(policy);
  }

  /// Define the time interval used for the reconnect schedule after the Quick Reconnect phase.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink to learn about the Quick Reconnect phase
  /// and the reconnect schedule for the policy defined by this method.
  ///
  /// \param seconds   a time interval expressed in seconds used for reconnect schedule; the default is \c 20
  setReconnectTimeInterval(int seconds) {
    clientImpl.setReconnectTimeInterval(seconds);
  }

  /// Define the maximum reconnect delay for the MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF policy.
  ///
  /// See \link MigratoryDataClient.setQuickReconnectInitialDelay() \endlink to learn how the value defined by this
  /// method is used.
  ///
  /// \param seconds   the maximum reconnect delay when using the MigratoryDataClient.TRUNCATED_EXPONENTIAL_BACKOFF
  ///                  policy; the default value is \c 360 seconds
  setReconnectMaxDelay(int seconds) {
    clientImpl.setReconnectMaxDelay(seconds);
  }

  /// Assigns an external token to enable client communication via external services while the client is offline.
  ///
  /// Presuming a client goes offline and disconnects from the MigratoryData cluster. Then plugins within the MigratoryData 
  /// server, leveraging MigratoryData's Presence API, can utilize this external token to communicate with the offline 
  /// client, via external services like push notifications or email.
  ///
  /// For instance, consider a plugin designed for Firebase Cloud Messaging (FCM), which requires an FCM token to send 
  /// push notifications to a client via FCM. Through this method, the client can provide the necessary FCM token to the plugin.
  /// 
  /// Similarly, consider a plugin designed for email services. To send emails to clients when they're offline, the 
  /// plugin requires their email addresses. Clients can provide their email addresses using this method.
  ///
  /// \param externalToken   a token (e.g. FCM token, email address, ...)
  setExternalToken(String externalToken) {
    clientImpl.setExternalToken(externalToken);
  }

  setSocketTimeoutSeconds(int seconds) {
    clientImpl.setSocketTimeoutSeconds(seconds);
  }

  void pause() {
    clientImpl.pause();
  }

  void resume() {
    clientImpl.resume();
  }

  void setHttpHeader(String header, String value,
      {bool preserveHeaderCase = true}) {
    clientImpl.setHttpHeader(header, value, preserveHeaderCase);
  }
}
