import 'dart:convert';
import 'dart:core';
import 'dart:typed_data';

enum MessageType { SNAPSHOT, UPDATE, RECOVERED, HISTORICAL }

/// The quality of service (QoS) levels for MigratoryData messaging.
enum QoS {
  ///////
  /// The <code>QoS.STANDARD</code> should be used for noncritical messages which will not be included in the cache
  /// of the MigratoryData cluster. In this way, after a connection recovery, a client will not receive as part of
  /// the recovery process the messages with <code>QoS.STANDARD</code>.
  ///
  /// <p>
  ///     Note: This QoS level corresponds to the <code>at-most-once</code> delivery semantics.
  /// </p>
  ////
  STANDARD,

  ///////
  /// The <code>QoS.GUARANTEED</code> should be used for critical messages which will be included in the cache of
  /// the MigratoryData cluster. In this way, after a connection recovery, a client will receive as part of the
  /// recovery process the messages with <code>QoS.GUARANTEED</code>.
  ///
  /// <p>
  ///     Note: This QoS level corresponds to the <code>at-least-once</code> delivery semantics. Also, it might
  ///     correspond to the <code>exactly-once</code> delivery semantics provided that subscribers are responsible
  ///     for filtering out duplicate receptions, if it at all matters for the application. Typically, a small
  ///     buffer containing the identifiers of recently-received messages is sufficient for this task.
  /// </p>
  ////
  GUARANTEED
}

/// Represent a MigratoryData message.
class MigratoryDataMessage {
  String subject;
  Uint8List content;

  String closure = "";
  String replySubject = "";
  QoS qos = QoS.GUARANTEED;

  bool retained = true;
  int seq = 0;
  int epoch = 0;
  bool compression = false;
  MessageType messageType = MessageType.UPDATE;

  ///////
  /// Create a MigratoryDataMessage object.
  ///
  /// If a reply subject is attached to a message, the message acts as a request message. The clients which receive a
  /// request message will be able to reply by sending back one or more messages having as subject the reply subject.
  ///
  /// <p>
  ///     Note: If the reply subject is not already subscribed, it will be subscribed by the library implicitly, and it
  ///     can be reused for request/reply interactions (and even for receiving multiple replies to one request). When
  ///     it is not needed anymore, it should be unsubscribed explicitly.
  /// </p>
  ///
  /// \param subject        the subject of the message
  /// \param content        the content of the message
  /// \param closure        the closure of the message
  /// \param qos            the QoS level of the message; the default is <code>QoS.GUARANTEED</code>
  /// \param retained       indicate whether or not the message should be/was retained by the server; the default is <code>true</code>
  /// \param replySubject   the reply subject of the message
  /// \param compression     if true, the content of the message will be published in ZLIB-compressed format; if false
  ///                       no compression format will be used; the default value is false.
  ///  <p>
  ///      Note: If compression is enabled with this method but the size of the content of the message is smaller than
  ///      the size of compressed content, then the message will be sent uncompressed to save bandwidth and CPU cycles
  ///      with the decompression on the receiver side.
  ///  </p>

  MigratoryDataMessage(this.subject, this.content,
      {this.closure = "",
      this.qos = QoS.GUARANTEED,
      this.retained = true,
      this.replySubject = "",
      this.compression = false});

  ///////
  /// Get the subject of the message.
  ///
  /// \return   the subject of the message
  ////
  String getSubject() {
    return subject;
  }

  ///////
  /// Get the content of the message.
  ///
  /// \return   the content of the message
  ////
  Uint8List getContent() {
    return content;
  }

  ///////
  /// Get the closure of the message.
  ///
  /// \return   the closure of the message
  ////
  String getClosure() {
    return closure;
  }

  ///////
  /// Indicate whether or not the message should be/was retained by the server.
  ///
  /// \return  true if the message should be, or was, retained by the server
  ////
  bool isRetained() {
    return retained;
  }

  ///////
  /// Get the reply subject of the message.
  ///
  /// @return   the reply subject of the message
  ////
  String getReplySubject() {
    return replySubject;
  }

  ///////
  /// Get the QoS level of the message.
  ///
  /// \return the QoS level of the message
  ////
  QoS getQos() {
    return qos;
  }

  ///////
  /// Get the MessageType of the message.
  ///
  /// \return the message type of the message
  ////
  MessageType getMessageType() {
    return messageType;
  }

  ///////
  /// Indicate whether or not the message should be/was compressed.
  ///
  /// @return Return whether or not the message should be/was compressed. If the received message was compressed, the
  /// library automatically decompresses its content without any additional code.
  ////
  bool isCompressed() {
    return compression;
  }

  /// Return a string representation of the message.
  @override
  String toString() {
    return "[subject=$subject, content=${Utf8Decoder(allowMalformed: true).convert(content)}, closure=$closure, retained=$retained, qos=$qos, messageType=$messageType, seq=$seq, epoch=$epoch, replySubject=$replySubject, compressed=$compression]";
  }

  int getSeq() {
    return seq;
  }

  int getEpoch() {
    return epoch;
  }
}
