import 'dart:typed_data';

import '../logger/logger.dart';
import 'message.dart';
import 'push_protocol.dart';
import 'push_protocol_transport_codec.dart';

class PushDecoder {

  final PushProtocol PUSH_PROTOCOL = PushProtocol.instance;
  final PushProtocolTransportCodec TRANSPORT_CODEC =
      PushProtocolTransportCodec.instance;

  List<Message> parseWebsocketFrame(Uint8List buf, Logger logger) {
    int end = buf.length;

    if (end == -1) {
      return List<Message>.empty();
    }

    int start = 0;
    List<Message> messages = List<Message>.empty(growable: true);
    for (;;) {
      if (start >= end) {
        return messages;
      }

      if (buf.elementAt(start) == PUSH_PROTOCOL.getUpdateLimit()) {
        start++;
      } else {
        for (;;) {
          int i = indexOf(buf, start, end, PUSH_PROTOCOL.getMessageLimit());
          if (i == -1) {
            break;
          }

          Map<Header, Object> headers =
              parsePushMessage(buf, start + 1, i, logger);
          if (headers.isNotEmpty) {
            Message msg = Message(
                PUSH_PROTOCOL.getOperation(buf.elementAt(start)), headers);
            messages.add(msg);
          }
          start = i + 1;
        }
      }
    }
  }

  int indexOf(Uint8List buf, int start, int end, int b) {
    for (int i = start; i < end; i++) {
      if (buf.elementAt(i) == b) {
        return i;
      }
    }
    return -1;
  }

  Map<Header, Object> parsePushMessage(
      Uint8List buf, int start, int end, Logger logger) {
    Map<Header, Object> headers = {};
    for (;;) {
      if (start >= end) {
        break;
      }
      int hdr = buf.elementAt(start);
      int hdrEnd = indexOf(buf, start + 1, end, PUSH_PROTOCOL.getHeaderEnd());
      if (hdrEnd == -1) {
        if (logger.isTraceEnabled()) {
          logger.trace(
              "Received an invalid message: header end missing - message ignored, header position: $start, $buf");
        }
        return {};
      }
      Header header = PUSH_PROTOCOL.getHeader(hdr);
      if (header == Header.NULL) {
        if (logger.isTraceEnabled()) {
          logger.trace(
              "Received an unknown header - header ignored, header position: $buf");
        }
        start = hdrEnd + 1;
        continue;
      }

      start++;

      switch (PUSH_PROTOCOL.getHeaderType(header)) {
        case HeaderType.INT:
          headers[header] = TRANSPORT_CODEC.bytesToInteger(
              buf, start, hdrEnd - start, PUSH_PROTOCOL);
          break;
        case HeaderType.STRING:
          Uint8List bytesVal = TRANSPORT_CODEC.unescape(
              buf, start, hdrEnd - start, PUSH_PROTOCOL);
          headers[header] = String.fromCharCodes(bytesVal);
          break;
        case HeaderType.BYTES:
          headers[header] = TRANSPORT_CODEC.unescape(
              buf, start, hdrEnd - start, PUSH_PROTOCOL);
          break;
        case HeaderType.UNESCAPED_BYTES:
          headers[header] = buf.sublist(start, hdrEnd);
          break;
        case HeaderType.NULL:
          break;
      }
      start = hdrEnd + 1;
    }
    return headers;
  }
}
