import 'dart:typed_data';
import 'dart:convert';

import '../core/compression/compression_stub.dart'
    if (dart.library.html) '../core/compression/compression_html.dart'
    if (dart.library.io) '../core/compression/compression_io.dart';

import '../core/subject.dart';
import '../migratorydata_message.dart';
import '../utils/monotonic_id.dart';
import 'buffer.dart';
import 'push_protocol.dart';
import 'push_protocol_transport_codec.dart';

class PushEncoder {
  final PushProtocol PUSH_PROTOCOL = PushProtocol.instance;
  final PushProtocolTransportCodec TRANSPORT_CODEC =
      PushProtocolTransportCodec.instance;
  HttpEncoding encoding = HttpEncoding.WEB_SOCKET;

  final compressionHelper = CompressionHelperImpl();

  void encodeConnect(
      Buffer buf,
      String entitlementInfo,
      Uint8List sessionType,
      Uint8List userAgent,
      int version,
      String externalToken,
      MonotonicId monotonicId) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.CONNECT));

    if (entitlementInfo.isNotEmpty) {
      addHeader(
          buf,
          PUSH_PROTOCOL.getHeaderByte(Header.ENTITLEMENT_TOKEN),
          TRANSPORT_CODEC
              .escape(Uint8List.fromList(utf8.encode(entitlementInfo))));
    }

    if (externalToken.isNotEmpty) {
      addHeader(
          buf,
          PUSH_PROTOCOL.getHeaderByte(Header.EXTERNAL_TOKEN),
          TRANSPORT_CODEC
              .escape(Uint8List.fromList(utf8.encode(externalToken))));

      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.MONOTONIC_ID),
          TRANSPORT_CODEC.integerToBytes(monotonicId.get()));
    }

    addHeader(
        buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_TYPE), sessionType);
    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.USER_AGENT), userAgent);
    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.VERSION),
        TRANSPORT_CODEC.integerToBytes(version));
    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodeSubscribe(Buffer buf, Subject subject, int sessionId) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.SUBSCRIBE));

    addHeader(
        buf,
        PUSH_PROTOCOL.getHeaderByte(Header.SUBJECT),
        TRANSPORT_CODEC
            .escape(Uint8List.fromList(utf8.encode(subject.getSubject()))));

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    switch (subject.getSubscribeType()) {
      case SubscribeType.NORMAL:
        break;
      case SubscribeType.HISTORY:
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.HISTORICAL_MESSAGES),
            TRANSPORT_CODEC.integerToBytes(subject.getHistory()));
        break;
      case SubscribeType.RECOVERY:
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SEQ_ID),
            TRANSPORT_CODEC.integerToBytes(subject.getSeqId()));
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SEQ),
            TRANSPORT_CODEC.integerToBytes((subject.getSeq() + 1)));
        break;
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodeUnsubscribe(Buffer buf, int sessionId, Subject sbj) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.UNSUBSCRIBE));

    addHeader(
        buf,
        PUSH_PROTOCOL.getHeaderByte(Header.SUBJECT),
        TRANSPORT_CODEC
            .escape(Uint8List.fromList(utf8.encode(sbj.getSubject()))));

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodePing(Buffer buf, int sessionId) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.PING));

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodeAuthorizationTokenUpdate(Buffer buf, String token, int sessionId) {
    buf.put(
        PUSH_PROTOCOL.getOperationByte(Operation.AUTHORIZATION_TOKEN_UPDATE));

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    if (token.isNotEmpty) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENTITLEMENT_TOKEN),
          TRANSPORT_CODEC.escape(Uint8List.fromList(utf8.encode(token))));
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodeDisconnect(Buffer buf, int sessionId) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.DISCONNECT));

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void encodePublish(Buffer buf, MigratoryDataMessage message, int sessionId) {
    buf.put(PUSH_PROTOCOL.getOperationByte(Operation.CLIENT_PUBLISH));

    addHeader(
        buf,
        PUSH_PROTOCOL.getHeaderByte(Header.SUBJECT),
        TRANSPORT_CODEC
            .escape(Uint8List.fromList(utf8.encode(message.getSubject()))));

    bool contentIsCompressed = false;
    if (message.isCompressed()) {
      String encodedCompressedData =
          compressionHelper.zLibCompressAndEncodeBase64(message.getContent());

      if (encodedCompressedData.length < message.getContent().length) {
        addHeader(
            buf,
            PUSH_PROTOCOL.getHeaderByte(Header.DATA),
            TRANSPORT_CODEC.escape(
                Uint8List.fromList(utf8.encode(encodedCompressedData))));
        contentIsCompressed = true;
      } else {
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.DATA),
            TRANSPORT_CODEC.escape(message.getContent()));
      }
    } else {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.DATA),
          TRANSPORT_CODEC.escape(message.getContent()));
    }

    if (message.getReplySubject().isNotEmpty) {
      addHeader(
          buf,
          PUSH_PROTOCOL.getHeaderByte(Header.REPLY_SUBJECT),
          TRANSPORT_CODEC.escape(
              Uint8List.fromList(utf8.encode(message.getReplySubject()))));
    }

    if (message.getClosure().isNotEmpty) {
      addHeader(
          buf,
          PUSH_PROTOCOL.getHeaderByte(Header.CLOSURE),
          TRANSPORT_CODEC
              .escape(Uint8List.fromList(utf8.encode(message.getClosure()))));
    }

    if (sessionId >= 0) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.SESSION_ID),
          TRANSPORT_CODEC.integerToBytes(sessionId));
    }

    switch (message.getQos()) {
      case QoS.GUARANTEED:
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.PUBLICATION_QOS),
            TRANSPORT_CODEC.integerToBytes(QoS.GUARANTEED.index));
        break;
      case QoS.STANDARD:
        addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.PUBLICATION_QOS),
            TRANSPORT_CODEC.integerToBytes(QoS.STANDARD.index));
        break;
    }

    if (message.isRetained()) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.PUBLICATION_RETAINED),
          TRANSPORT_CODEC.integerToBytes(1));
    } else {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.PUBLICATION_RETAINED),
          TRANSPORT_CODEC.integerToBytes(0));
    }

    if (contentIsCompressed) {
      addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.COMPRESSION),
          TRANSPORT_CODEC.integerToBytes(1));
    }

    addHeader(buf, PUSH_PROTOCOL.getHeaderByte(Header.ENCODING),
        TRANSPORT_CODEC.integerToBytes(encoding.code));

    buf.put(PUSH_PROTOCOL.getMessageLimit());
  }

  void addHeader(Buffer buf, int header, Uint8List content) {
    buf.put(header);
    buf.putArray(content);
    buf.put(PUSH_PROTOCOL.getHeaderEnd());
  }
}
